

Type TExplosion Extends TProp
	Method New()
		class = "explosion"
		solid = False
		radius = 1.0
		animSpeed = 3
	End Method
	Method Update()
		yaw = -ATan2((gPlayer.z - gPlayer.camSin) - z, (gPlayer.x - gPlayer.camCos) - x) + 90.0
		anim :+ 1
		If anim > animSpeed Then
			anim = 0
			If TTexture(MapValueForKey(gTextures, "explosion_" + (frame + 1))) <> Null Then
				frame :+ 1
			Else
				frame = 0
				RemoveEntity(Self)
			EndIf
		EndIf
		texture = TTexture(MapValueForKey(gTextures, "explosion_" + frame))
	End Method
	Method Render()
		glBindTexture GL_TEXTURE_2D, texture.id
		glMatrixMode(GL_MODELVIEW)
		glPushMatrix()
		glTranslatef x, y, z
		glRotatef yaw, 0, 1, 0
		glScalef 2.0, 2.0, 2.0
		
		glCallList(gDisplayLists[1])
		
		glLoadIdentity()
		glMatrixMode(GL_MODELVIEW)
		glPopMatrix()
		glBindTexture GL_TEXTURE_2D, Null
	End Method
End Type

Function Explode(x:Float, y:Float, z:Float, radius:Float, damage:Int = 20, direction:Float = 0.0, owner:TEntity = Null)
	Local ex:TExplosion = New TExplosion
	ex.x=x - GetCos(direction) * 0.01
	ex.y=y
	ex.z=z - GetSin(direction) * 0.01
	ex.radius = radius
	AddEntity(ex)
	For Local m:TMonster = EachIn gMonsters
		If owner = m Or m.state = 2 Then Continue
		Local dist:Float = Distance(x, z, m.x, m.z)
		If dist <= radius Then
			If m.active = False Then m.active = True
			m.Hurt(damage, False)	
		EndIf
	Next
	If Distance(x, z, gPlayer.x, gPlayer.z) - gPlayer.radius <= radius Then
		gPlayer.health :- Abs(damage)
		FlashScreen(255, 0, 0, 1.0)
	EndIf
	Play3DSound("explosion", x, z)
End Function

Type TMedkit Extends TProp
	Field amount:Int = 25
	Method New()
		class = "prop"
		solid = False
	End Method
	Method OnPickup()
		If (gPlayer.health < 100 And gGameDifficulty = DIFFICULTY_NORMAL) Or (gPlayer.health < 200 And gGameDifficulty >= DIFFICULTY_EASY) Then
			FlashScreen(255, 255, 255, 0.75)
			gPlayer.health :+ amount
			If gGameDifficulty >= DIFFICULTY_EASY Then
				gPlayer.health :+ amount
				If gGameDifficulty = DIFFICULTY_EASIEST Then
					gPlayer.health :+ Ceil(amount / 2.0)
				EndIf
			EndIf
			If gPlayer.health > 100 And gGameDifficulty = DIFFICULTY_NORMAL Then
				gPlayer.health = 100
			EndIf
			If gPlayer.health > 200 And gGameDifficulty >= DIFFICULTY_EASY Then
				gPlayer.health = 200
			EndIf
			RemoveEntity(Self)
			SetChannelVolume(PlaySound(TSound(MapValueForKey(gSounds, "pickup"))),sSoundVolume)
		EndIf
	End Method
EndType

Type TArmorPickup Extends TProp
	Field r:Byte = 0
	Field g:Byte = 0
	Field b:Byte = 0
	Field armorType:Byte = 0
	Field armor:Byte = 50
	Field obituary:String = ""
	Method New()
		class = "prop"
		solid = False
	End Method
	Method OnPickup()
		FlashScreen(r, g, b, 1.0)
		gPlayer.armor = armor
		gPlayer.armorType = armorType
		gMessage = obituary
		gMessageTimer = 50
		RemoveEntity(Self)
		SetChannelVolume(PlaySound(TSound(MapValueForKey(gSounds, "armor"))),sSoundVolume)
	End Method
EndType

Type TAmmoPickup Extends TProp
	Field weapon:TWeapon
	Field amount:Byte = 5
	Field obituary:String = ""
	
	Field flag:Int = False
	Method New()
		class = "prop"
		solid = False
	End Method
	Method OnPickup()
		
		flag = False
		Select weapon.ammoType
			Case AMMO_EGG
				If gAmmoTypes[AMMO_EGG] >= 256 Then flag = True
			Case AMMO_GRENADE
				If gAmmoTypes[AMMO_GRENADE] >= 100 Then flag = True
			Case AMMO_PLASMA
				If gAmmoTypes[AMMO_PLASMA] >= 500 Then flag = True
		EndSelect
		If flag = False Then 
			FlashScreen(255, 0, 255, 0.75)
			gAmmoTypes[weapon.ammoType] :+ amount
			If gGameDifficulty = DIFFICULTY_EASY Then
				gAmmoTypes[weapon.ammoType] :+ amount
			Else If gGameDifficulty = DIFFICULTY_EASIEST Then
				gAmmoTypes[weapon.ammoType] :+ amount
				gAmmoTypes[weapon.ammoType] :+ Ceil(amount / 2.0)
			EndIf
			gMessage = obituary
			gMessageTimer = 50
			RemoveEntity(Self)
			SetChannelVolume(PlaySound(TSound(MapValueForKey(gSounds, "pickup"))),sSoundVolume)

		EndIf
	End Method
EndType

Type TKeyCard Extends TProp
	Field color:Byte = 0
	Method New()
		class = "prop"
		solid = False
	End Method
	Method OnPickup()
		Select color
			Case 0
				FlashScreen(0, 0, 255, 1.0)
			Case 1
				FlashScreen(128, 0, 0, 1.0)
			Case 2
				FlashScreen(255, 255, 0, 1.0)
			Case 3
				FlashScreen(128, 128, 128, 1.0)
		End Select
		gKeycards[color] = True
		RemoveEntity(Self)
		Play3DSound("key", x, z)
	End Method
End Type

Type TDoor Extends TEntity
	Field orientation:Byte = 0
	Field newX:Float = 0.0
	Field newZ:Float = 0.0
	Field orgX:Float = 0.0
	Field orgZ:Float = 0.0
	Field newY:Float = 0.0
	Field orgY:Float = 0.0
	Field state:Byte = 0
	Field speed:Float = 0.05
	Field locked:Int = False
	Field key:Byte = 0
	Field d:Byte = False
	Field s:Byte = True
	Field e:TEntity
	
	Method New()
		class = "wall:door"
		planes = New TPlane[4]
	End Method
	Method Update()
		planes[0]    = MakePlane(MakeVector(x-1.0,z-1.0), MakeVector(x+1.0,z-1.0))
		planes[1]    = MakePlane(MakeVector(x-1.0,z+1.0), MakeVector(x+1.0,z+1.0))
		planes[2]    = MakePlane(MakeVector(x+1.0,z-1.0), MakeVector(x+1.0,z+1.0))
		planes[3]    = MakePlane(MakeVector(x-1.0,z-1.0), MakeVector(x-1.0,z+1.0))
		If state = 1 Then
			s = True
			If newX > x Then
				x = Min(x + speed, newX)
			ElseIf newX < x Then
				x = Max(x - speed, newX)
			EndIf
			If newZ > z Then
				z = Min(z + speed, newZ)
			ElseIf newZ < z Then
				z = Max(z - speed, newZ)
			EndIf
			If newY > y Then
				y = Min(y + speed, newY)
			ElseIf newY < y
				y = Max(y - speed, newY)
			EndIf
			
			If gPlayer.x + gPlayer.radius > x-1.0 And gPlayer.x - gPlayer.radius < x+1.0 And gPlayer.z + gPlayer.radius > z-1.0 And gPlayer.z - gPlayer.radius < z+1.0 Then
				If orientation = 0 Then
					newX = orgX + 1.8
				ElseIf orientation = 1
					newZ = orgZ + 1.8
				Else
					newY = orgY + 1.8
				EndIf
			EndIf
			For Local m:TMonster = EachIn gMonsters
				If m.class = "monster" Then
					If m.x+m.radius > x-1.0 And m.x - m.radius < x+1.0 And m.z+m.radius > z-1.0 And m.z - m.radius < z+1.0 Then
						If orientation = 0 Then
							newX = orgX + 1.8

						ElseIf orientation = 1
							newZ = orgZ + 1.8

						Else
							newY = orgY +1.8

						EndIf
					EndIf
				EndIf
			Next
			
			If (newX = x And orientation = 0) Or (newZ = z And orientation = 1) Or (newY = y And orientation = 2) Then
				state = 2
				If (newX = orgX And orientation = 0) Or (newZ = orgZ And orientation = 1) Or (newY = orgY And orientation = 2) Then state = 0
			EndIf
		ElseIf state = 2 Then
			s = False
			If orientation = 2 Then
				
			EndIf
		Else
			s = True
		EndIf
	End Method
	Method Render()
		If Distance(x, z, gPlayer.x, gPlayer.z) < 48.0 Then
			glBindTexture GL_TEXTURE_2D, texture.id
			glMatrixMode(GL_MODELVIEW)
			glPushMatrix()
			glTranslatef x, y, z
			
			If orientation = 0 Then glCallList(gDisplayLists[2])
			If orientation = 1 Then glCallList(gDisplayLists[3])
			If orientation = 2 Then
				If d = False Then glCallList(gDisplayLists[2])
				If d = True Then glCallList(gDisplayLists[3])
			EndIf
			
			glLoadIdentity()
			glMatrixMode(GL_MODELVIEW)
			glPopMatrix()
			glBindTexture GL_TEXTURE_2D, Null
		EndIf
	End Method
	Method Toggle()
		If state <> 1 And locked = False Then
			Play3DSound("opendoor", x, z)
			If orientation = 0 Then
				If state = 0 Then
					newX = orgX + 1.8
					state = 1
				ElseIf state = 2 Then
					newX = orgX
					state = 1
				EndIf
			ElseIf orientation = 1 Then
				If state = 0 Then
					newZ = orgZ + 1.8
					state = 1
				ElseIf state = 2 Then
					newZ = orgZ
					state = 1
				EndIf
			ElseIf orientation = 2 Then
				If state = 0 Then
					newY = orgY + 1.8
					state = 1
				ElseIf state = 2 Then
					newY = orgY
					state = 1
				EndIf
			EndIf
		EndIf
		If state <> 1 And locked = True And key > 0 Then
			If gKeycards[key - 1] = True Then
				Play3DSound("opendoor", x, z)
				If orientation = 0 Then
					If state = 0 Then
						newX = orgX + 1.8
						state = 1
					ElseIf state = 2 Then
						newX = orgX
						state = 1
					EndIf
				ElseIf orientation = 1 Then
					If state = 0 Then
						newZ = orgZ + 1.8
						state = 1
					ElseIf state = 2 Then
						newZ = orgZ
						state = 1
					EndIf
				ElseIf orientation = 2 Then
					If state = 0 Then
						newY = orgY + 1.8
						state = 1
					ElseIf state = 2 Then
						newY = orgY
						state = 1
					EndIf
				EndIf
			Else
				gMessageTimer = 100
				Select key
					Case 1
						gMessage = "YOU NEED A BLUE KEYCARD"
					Case 2
						gMessage = "YOU NEED A BROWN KEYCARD"
					Case 3
						gMessage = "YOU NEED A YELLOW KEYCARD"
					Case 4
						gMessage = "YOU NEED A GRAY KEYCARD"
				EndSelect
			EndIf
		EndIf
		If locked = True And key = 0 Then
			gMessage = "THIS DOOR IS OPENED ELSEWHERE"
			gMessageTimer = 100
		EndIf
	End Method
	Method Open()
		If state <> 1 Then
			If orientation = 0 Then
				newX = orgX + 1.8
				state = 1
			ElseIf orientation = 1 Then
				newZ = orgZ + 1.8
				state = 1
			ElseIf orientation = 2
				newY = orgY + 1.8
				state = 1
			EndIf
		EndIf
	End Method
	Method Close()
		If state <> 1 Then
			If orientation = 0 Then
				newX = orgX
				state = 1
			ElseIf orientation = 1 Then
				newZ = orgZ
				state = 1
			ElseIf orientation = 2 Then
				newY = orgY
				state = 1
			EndIf
		EndIf
	End Method
End Type

Type TFlatWall Extends TEntity
	Field orientation:Byte = 0
	Field solid:Byte = True
	Method New()
		class = "wall:flat"
		planes = New TPlane[4]
	End Method
	Method Update()
		If planes[0] = Null Then
			planes[0]    = MakePlane(MakeVector(x-1.0,z-1.0), MakeVector(x+1.0,z-1.0))
			planes[1]    = MakePlane(MakeVector(x-1.0,z+1.0), MakeVector(x+1.0,z+1.0))
			planes[2]    = MakePlane(MakeVector(x+1.0,z-1.0), MakeVector(x+1.0,z+1.0))
			planes[3]    = MakePlane(MakeVector(x-1.0,z-1.0), MakeVector(x-1.0,z+1.0))
		EndIf
	End Method
	Method Render()
		If Distance(x, z, gPlayer.x, gPlayer.z) < 48.0 Then
			glBindTexture GL_TEXTURE_2D, texture.id
			glMatrixMode(GL_MODELVIEW)
			glPushMatrix()
			glTranslatef x, y, z
			
			If orientation = 0 Then glCallList(gDisplayLists[2])
			If orientation = 1 Then glCallList(gDisplayLists[3])
			
			glLoadIdentity()
			glMatrixMode(GL_MODELVIEW)
			glPopMatrix()
			glBindTexture GL_TEXTURE_2D, Null
		EndIf
	End Method
End Type

Type TSecretWall Extends TWall
	Field activated:Byte = False
	Field done:Byte = False
	Field orgX:Float = 3.14
	Field orgZ:Float = 3.14
	Field locked:Byte = False
	Method New()
		class = "secretwall"
		planes = New TPlane[4]
	End Method
	Method Update()
		If planes[0] = Null Then
			planes[0]    = MakePlane(MakeVector(x-1.0,z-1.0), MakeVector(x+1.0,z-1.0))
			planes[1]    = MakePlane(MakeVector(x-1.0,z+1.0), MakeVector(x+1.0,z+1.0))
			planes[2]    = MakePlane(MakeVector(x+1.0,z-1.0), MakeVector(x+1.0,z+1.0))
			planes[3]    = MakePlane(MakeVector(x-1.0,z-1.0), MakeVector(x-1.0,z+1.0))
		EndIf
		If orgX = 3.14 Or orgZ = 3.14 Then
			orgX = x
			orgZ = z
		EndIf
		If activated = True And done = False Then
			If locked = True Then
				RemoveEntity(Self)
				ListRemove gSecretWalls, Self
			Else
				Select id
					Case 0 'Moves Right
						x :+ 0.125
						If x > orgX + 4.0 Then
							done = True
							x = orgX + 4.0
						EndIf
					Case 1 'Move Up
						z :- 0.125
						If z < orgZ - 4.0 Then
							done = True
							z = orgZ - 4.0
						EndIf
					Case 2 'Move Left
						x :- 0.125
						If x < orgX - 4.0 Then
							done = True
							x = orgX - 4.0
						EndIf
					Case 3 ' Move Down
						z :+ 0.125
						If z > orgZ + 4.0 Then
							done = True
							z = orgZ + 4.0
						EndIf
				End Select
			EndIf
		EndIf
	End Method
End Type

Type TSwitch Extends TEntity
	Field state:Byte = False
	Method New()
		class = "wall:switch"
		planes = New TPlane[4]
	End Method
	Method Update()
		If planes[0] = Null Then
			planes[0]    = MakePlane(MakeVector(x-1.0,z-1.0), MakeVector(x+1.0,z-1.0))
			planes[1]    = MakePlane(MakeVector(x-1.0,z+1.0), MakeVector(x+1.0,z+1.0))
			planes[2]    = MakePlane(MakeVector(x+1.0,z-1.0), MakeVector(x+1.0,z+1.0))
			planes[3]    = MakePlane(MakeVector(x-1.0,z-1.0), MakeVector(x-1.0,z+1.0))
		EndIf
	End Method
	Method Render()
		If state = True Then glBindTexture GL_TEXTURE_2D, TTexture(MapValueForKey(gTextures, texture.name + "_on")).id
		If state = False Then glBindTexture GL_TEXTURE_2D, texture.id
		glMatrixMode(GL_MODELVIEW)
		glPushMatrix()
		glTranslatef x, y, z
		
		glCallList(gDisplayLists[0])
		
		glLoadIdentity()
		glMatrixMode(GL_MODELVIEW)
		glPopMatrix()
		glBindTexture GL_TEXTURE_2D, Null
	End Method
	Method Toggle()
		state = Not state
		If state = True Then
			ActivateAllWithId(id)
		EndIf
		SetChannelVolume(PlaySound(TSound(MapValueForKey(gSounds, "switch"))),sSoundVolume)
	End Method
End Type

Type TTrigger Extends TEntity
	Field isExit:Byte = False
	
	Field prop:TProp
	Method New()
		class = "trigger"
	End Method
	Method Update()
		If gPlayer.x+gPlayer.radius >= x-1.0 And gPlayer.x-gPlayer.radius <= x+1.0 And gPlayer.z+gPlayer.radius >= z-1.0 And gPlayer.z-gPlayer.radius <= z+1.0 And gPlayer.state <> 2 Then
			If isExit = False Then
				ActivateAllWithId(id)
				If id = 255 Then
					gSecretStat :+ 1
				EndIf
				RemoveEntity(Self)
			Else
				gPlayer.state = 2
				
				prop = New TProp
				prop.x = gPlayer.x
				prop.z = gPlayer.z
				prop.y = 0.0
				prop.baseTextureName = "SEGAN_VICTORY"
				prop.solid = True
				prop.texture = TTexture(MapValueForKey(gTextures, "SEGAN_VICTORY_0"))
				prop.animated = True
				prop.loopAnim = False
				AddEntity(prop)
				
				StopChannel gMusicChannel
				gMusicChannel = PlaySound(gVictoryMusic)
				
				If gMapNumber <> 9 Then gMapNumber :+ 1
				If id > 0 Then gMapNumber = id
				If gMapNumber > 10 Then
					gMapNumber = 10
				EndIf
				
				'2192 992
			EndIf
		EndIf
	End Method
EndType

Function ActivateAllWithId(id:Byte)
	For Local e:TEntity = EachIn gEntities
		If Instr(e.class,"door") > 0 And e.id = id Then
			Local d:TDoor = TDoor(e)
			If d.key = 0 Then
				If d.state = 0 Then Play3DSound("opendoor", e.x, e.z)
				d.Open()
			EndIf
		ElseIf e.class = "secretwall" And e.id = id Then
			Local sw:TSecretWall = TSecretWall(e)
			If sw.locked = True Then
				sw.activated = True
			EndIf
		EndIf
	Next
End Function

Type TWeaponPickup Extends TProp
	Method New()
		class = "prop"
		solid = False
	End Method
	Method OnPickup()
		FlashScreen(0, 255, 0, 1.0)
		If texture.name = "chickencannon" Then
			gNewWeapon = gWeaponChicken
			gAmmoTypes[AMMO_EGG] :+ 50
		ElseIf texture.name = "grenadelauncher" Then
			gNewWeapon = gWeaponGrenade
			gAmmoTypes[AMMO_GRENADE] :+ 5
		ElseIf texture.name = "doublegrenadelauncher" Then
			gNewWeapon = gWeaponDoubleGrenade
			gAmmoTypes[AMMO_GRENADE] :+ 10
		ElseIf texture.name = "parusu" Then
			gNewWeapon = gWeaponParusu
			gAmmoTypes[AMMO_PLASMA] :+ 100
		ElseIf texture.name = "signofmadness" Then
			gNewWeapon = gWeaponSign
		ElseIf texture.name = "airhorn" Then
			gNewWeapon = gWeaponAirhorn
		EndIf
		gNewWeapon.equipped = True
		gMessage = gNewWeapon.obituary
		gMessageTimer = 100
		gNewWeapon.frame = 0
		gWeaponChange = 1
		gWeaponOffsetY = 0
		gWeaponBob = 0
		RemoveEntity( Self)
		SetChannelVolume(PlaySound(TSound(MapValueForKey(gSounds, "weapon"))),sSoundVolume)
	End Method
End Type

Type TSickle Extends TProjectile
	Field anim:Int = 0
	Field timer:Int = 0
	Field sound:TChannel
	Method New()
		class = "projectile"
		friendly = True
		radius = 0.25
		texture = New TTexture
		texture.name = "sickle_"
		sound = PlaySound(TSound(MapValueForKey(gSounds, "sickle_looped")))
	End Method
	Method Update()
		SetChannelVolume sound, sSoundVolume
		yaw = -ATan2((gPlayer.z - gPlayer.camSin) - z, (gPlayer.x - gPlayer.camCos) - x) + 90.0
		If speed < 0.0 Then direction = ATan2((gPlayer.z - z), (gPlayer.x - x)) + 180.0
		anim :+ 1
		timer :+ 1
		If anim > 2 Then
			anim = 0
			frame :+ 1
			If frame > 3 Then
				frame = 0
			EndIf
		EndIf
		speed :- 0.02
		If CheckWall(x, z, radius, Self) = True And speed > 0.0 Then
			speed = -speed
		EndIf
		dx = GetCos(direction) * speed
		dz = GetSin(direction) * speed
		x :+ dx
		z :+ dz
		If (Distance(gPlayer.x, gPlayer.z, x, z)-radius-gPlayer.radius <= 0.0 And speed < 0.0) Or timer > 200 Then
			RemoveEntity(Self)
			StopChannel sound
			sound = PlaySound(TSound(MapValueForKey(gSounds, "sickle_return")))
			SetChannelVolume sound, sSoundVolume
			If gWeapon = gWeaponSickle Then
				gPlayer.canShoot = True
				gWeapon.frame = 0
			EndIf
			gSicklePresent = False
		EndIf
	End Method
	Method Render()
		glBindTexture GL_TEXTURE_2D, TTexture(MapValueForKey(gTextures, texture.name + frame)).id
		glMatrixMode(GL_MODELVIEW)
		glPushMatrix()
		glTranslatef x, y, z
		glRotatef yaw, 0, 1, 0
		glScalef 0.75, 0.75, 0.75
		
		glCallList(gDisplayLists[1])
		
		glLoadIdentity()
		glMatrixMode(GL_MODELVIEW)
		glPopMatrix()
		glBindTexture GL_TEXTURE_2D, Null
	End Method
End Type

Type TEgg Extends TProjectile
	Field i:Int = 0
	Field chicken:TChicken
	Method New()
		class = "projectile"
		friendly = True
		radius = 0.1
		texture = TTexture(MapValueForKey(gTextures, "egg"))
		speed = 0.4
		damage = 5
		maxFrames = 1
	End Method
	Method Update()
		A_Projectile()
	End Method
	Method OnDestroy()
		For i = 0 To 4 
			AddParticle(x, y, z, (RndFloat()*0.25)-0.125, 0.2, (RndFloat()*0.25)-0.125, 0.025, 20, TTexture(MapValueForKey(gTextures, "particles_0")))
		Next
		If Rand(0, 30) = 2 Then
			chicken = New TChicken
			chicken.x = x - GetCos(direction)
			chicken.z = z - GetSin(direction)
			AddEntity(chicken)
			Play3DSound("chicken", x, z)
		EndIf
	End Method
	Method Render()
		glBindTexture GL_TEXTURE_2D, texture.id
		glMatrixMode(GL_MODELVIEW)
		glPushMatrix()
		glTranslatef x, y, z
		glRotatef yaw, 0, 1, 0
		
		glCallList(gDisplayLists[1])
		
		glLoadIdentity()
		glMatrixMode(GL_MODELVIEW)
		glPopMatrix()
		glBindTexture GL_TEXTURE_2D, Null
	End Method
EndType

Type TGrenade Extends TProjectile
	Field forward:Float = 0.25
	Field velocity:Float = 0.0
	Field anim:Int = 0
	Field timer:Int = 0
	
	Method New()
		class = "projectile"
		friendly = True
		radius = 0.15
		texture = New TTexture
		texture.name = "grenade_"
		damage = 25
	End Method
	Method Update()
		yaw = -ATan2((gPlayer.z - gPlayer.camSin) - z, (gPlayer.x - gPlayer.camCos) - x) + 90.0
		anim :+ 1
		If anim > 5 Then
			anim = 0
			frame = Not frame
		EndIf
		timer :+ 1
		velocity :- 0.005
		If y < -0.8 Then
			y = -0.8
			velocity = -velocity * 0.75
			forward :* 0.9
		EndIf
		If Abs(velocity) < 0.005 Then
			velocity = 0.0
			forward :* 0.9
		EndIf
		
		
		dx = GetCos(direction) * forward
		dz = GetSin(direction) * forward
		
		If CheckWall(x + dx, z + dz, radius, Self) = True Then
			'forward = -forward * 0.70
			direction :+ 90.0
			forward :* 0.9
		EndIf
		
		TryMove(x, z, dx, dz, radius, Self, False, False)
		
		x :+ dx
		y :+ velocity
		z :+ dz
		
		timer :+ 1
		If timer > 200 Then
			Explode(x, y + 0.5, z, 4.0, 20, direction, owner)
			RemoveEntity(Self)
		EndIf
	End Method
End Type

Type TFireball Extends TProjectile
	Method New()
		class = "projectile"
		friendly = False
		radius = 0.25
		texture = New TTexture
		texture.name = "FIREBALL_"
		speed = 0.45
		damage = 20
	End Method
	Method Update()
		A_Projectile()
	End Method
EndType

Type THomingBall Extends TProjectile
	Method New()
		class = "projectile"
		friendly = False
		radius = 0.25
		texture = New TTexture
		texture.name = "PLASMABALL_"
		speed = 0.65
		damage = 10
	End Method
	Method Update()
		A_Projectile()
		If y > 0.0 Then y :- 0.01
	End Method
End Type

Type TCacoBall Extends TProjectile
	Method New()
		class = "projectile"
		friendly = False
		radius = 0.25
		texture = New TTexture
		texture.name = "CACOBALL_"
		speed = 0.3
		damage = 20
	End Method
	Method Update()
		A_Projectile()
		For Local m:TMonster = EachIn gMonsters
			If m.class = "corpse" And Distance(m.x, m.z, x, z)-4.0-m.radius <= 0.0 And m <> owner Then
				If m.state = 2 Then m.Revive()
			EndIf
		Next
	End Method
	Method OnDestroy()
		For Local i:Int = 0 To 10
			AddParticle(x, y, z, (RndFloat()*1.0)-0.5, 0.2, (RndFloat()*1.0)-0.5, 0.025, 20, TTexture(MapValueForKey(gTextures, "particles_2")))
		Next
	EndMethod
EndType

Type TConveyor Extends TEntity
	Field direction:Float = 0.0
	Field speed:Float = 0.05
	Method New()
		class = "conveyor"
		radius = 1.1
	End Method
	Method Render()
		'You don't
	End Method
EndType

Type TWraith Extends TMonster
	Field pp:Float = 0.0
	Field wp:Float = 0.0
	Field ds0:Float = 0.0
	Field ds1:Float = 0.0
	Method New()
		class = "monster"
		radius = 0.4
		baseTextureName = "WRAITH"
		health = 20
		If gGameDifficulty = DIFFICULTY_EASIEST Then
			health = 10
		EndIf
		maxHealth = health
		walkAnimation = New Byte[4]
		walkAnimation[0] = 0
		walkAnimation[1] = 1
		walkAnimation[2] = 0
		walkAnimation[3] = 3
		animation = walkAnimation
		deathAnimation = New Byte[5]
		deathAnimation[0] = 4
		deathAnimation[1] = 5
		deathAnimation[2] = 6
		deathAnimation[3] = 7
		deathAnimation[4] = 8
		hurtFrame = 4
		speed = 0.1
		texture = New TTexture
		seeSound = TSound(MapValueForKey(gSounds, "wraith0"))
		painSound = TSound(MapValueForKey(gSounds, "wraith1"))
		deathSound = TSound(MapValueForKey(gSounds, "wraith3"))
		frame = 0
		anim = 0
		texture = TTexture(MapValueForKey(gTextures, "WRAITH0_0"))
		contactDamage = 2
		target = gPlayer
	End Method
	Method Activate()
		If foundWall = False Then
			active = True
			MonsterPlaySound(seeSound)
			angle = ATan2(target.z-z, target.x-x)
			If angle >= 360.0 Then angle :- 360.0
			If angle < 0.0 Then angle :+ 360.0
			gPlayer.hearRange = 30.0
		EndIf
	End Method
	Method Update2()
		A_WraithChase()
	End Method
End Type

Type TFireWraith Extends TMonster
	Field fireAnimation:Byte[]
	Field targetX:Float = 0.0
	Field targetZ:Float = 0.0
	Field standing:Int = False
	Field standTimer:Int = 0
	Field fireball:TFireball
	Field reactionTimer:Int = 0
	Field m:TMonster = Null
	Field volume:Float = 0.0
	Method New()
		class = "monster"
		radius = 0.5
		baseTextureName = "FIREWRAITH"
		health = 40
		If gGameDifficulty = DIFFICULTY_EASIEST Then
			health = 25
		EndIf
		maxHealth = health
		walkAnimation = New Byte[4]
		walkAnimation[0] = 0
		walkAnimation[1] = 1
		walkAnimation[2] = 0
		walkAnimation[3] = 3
		animation = walkAnimation
		deathAnimation = New Byte[4]
		deathAnimation[0] = 7
		deathAnimation[1] = 8
		deathAnimation[2] = 9
		deathAnimation[3] = 10
		fireAnimation = New Byte[2]
		fireAnimation[0] = 4
		fireAnimation[1] = 5
		hurtFrame = 6
		speed = 0.12
		texture = New TTexture
		seeSound = TSound(MapValueForKey(gSounds, "firewraith0"))
		painSound = TSound(MapValueForKey(gSounds, "firewraith1"))
		deathSound = TSound(MapValueForKey(gSounds, "firewraith2"))
		frame = 0
		anim = 0
		texture = TTexture(MapValueForKey(gTextures, "FIREWRAITH0_0"))
		contactDamage = 1
		target = gPlayer
		targetX = x + 1.0
		targetZ = z
		cwmf = True
		canConveyor = False
	End Method
	Method Activate()
		If foundWall = False Then
			active = True
			MonsterPlaySound(seeSound)
			angle = ATan2(target.z-z, target.x-x)
			If angle >= 360.0 Then angle :- 360.0
			If angle < 0.0 Then angle :+ 360.0
			targetX = x + 1.0
			targetZ = z
		EndIf
	End Method
	Method Update2()
		If state <> 2 And reactionTimer > 25 Then
			If standing = False Then
				animationRepeat = True
				If dist2p > 1.0 Then
					angle = ATan2(targetZ-z, targetX-x)
					If angle >= 360.0 Then angle :- 360.0
					If angle < 0.0 Then angle :+ 360.0
					If CheckWall(x+GetCos(angle)*speed, z+GetSin(angle)*speed, radius, Self, True) = True Then
						targetX = x + Rnd(-2.0, 2.0)
						targetZ = z + Rnd(-2.0, 2.0)
					EndIf
					dx = GetCos(angle) * speed
					dz = GetSin(angle) * speed
					For Local m:TMonster = EachIn gMonsters
						If m.class = "monster" And m <> Self Then
							If Distance(m.x, m.z, x+dx, z+dz)-radius-m.radius <= 1.0 Then
								targetX = X
								targetZ = Z
							EndIf
						EndIf
					Next
					If Distance(x, z, targetX, targetZ) < 0.25 Then
						targetX = x + Rnd(-3.0, 3.0)
						If Abs(targetX) < 1.0 Then targetX = 1.0 * Sgn(targetX)
						targetZ = z + Rnd(-3.0, 3.0)
						If Abs(targetZ) < 1.0 Then targetZ = 1.0 * Sgn(targetZ)
						standing = True
						angle = ATan2(target.z-z, target.x-x)
						animationRepeat = False
						anim = 0
						animation = fireAnimation
						frame = 0
						animSpeed = 5
					EndIf
					
				Else
					dx = 0.0
					dz = 0.0
				EndIf
			Else
				dx = 0
				dz = 0
				angle = ATan2(target.z-z, target.x-x)
				standTimer :+ 1
				If standTimer = 10 Then
					fireball = New TFireball
					fireball.x = x
					fireball.z = z
					fireball.direction = angle
					fireball.speed = 0.425
					fireball.damage = 15
					AddEntity(fireball)
					Play3DSound("fireball", x, z, 40.0)
				EndIf
				If standTimer > 50 Then
					standTimer = 0
					standing = False
					animation = walkAnimation
					frame = 0
					anim = 0
					animSpeed = 10
				EndIf
			EndIf
		Else
			reactionTimer :+ 1
		EndIf
	End Method
End Type

Type TDummkopf Extends TMonster
	Field reactionTimer:Int = 0
	Field stateTimer:Int = 0
	Field firing:Int = False
	Field fireTimer:Int = 0
	Field fireSpeed:Int = 10
	Field fireball:TFireball
	Field seeAnimation:Byte[]
	Field fireAnimation:Byte[]
	Field pawn:Byte = False
	Method New()
		class = "monster"
		radius = 0.5
		baseTextureName = "DUMMKOPF"
		health = 75
		If gGameDifficulty = DIFFICULTY_EASIEST Then
			health = 55
		EndIf
		maxHealth = health
		animation = New Byte[1]
		animation[0] = 0
		deathAnimation = New Byte[4]
		deathAnimation[0] = 6
		deathAnimation[1] = 7
		deathAnimation[2] = 8
		deathAnimation[3] = 9
		seeAnimation = New Byte[2]
		seeAnimation[0] = 1
		seeAnimation[1] = 2
		fireAnimation = New Byte[2]
		fireAnimation[0] = 3
		fireAnimation[1] = 4
		walkAnimation = New Byte[1]
		walkAnimation[0] = 2
		hurtFrame = 5
		speed = 0.0
		texture = New TTexture
		seeSound = TSound(MapValueForKey(gSounds, "dummkopf0"))
		painSound = TSound(MapValueForKey(gSounds, "dummkopf1"))
		deathSound = TSound(MapValueForKey(gSounds, "dummkopf2"))
		frame = 0
		anim = 0
		texture = TTexture(MapValueForKey(gTextures, "DUMMKOPF0_0"))
		contactDamage = 1
		target = gPlayer
	End Method
	Method Activate()
		If foundWall = False Then
			active = True
			MonsterPlaySound(seeSound)
			angle = ATan2(target.z-z, target.x-x)
			If angle >= 360.0 Then angle :- 360.0
			If angle < 0.0 Then angle :+ 360.0
			target = gPlayer
			animation = seeAnimation
			animationRepeat = False
		EndIf
	End Method
	Method OnDie()
		Local plasmaammo:TAmmoPickup = New TAmmoPickup
		plasmaammo.x = x
		plasmaammo.z = z
		plasmaammo.radius = 0.5
		plasmaammo.texture = TTexture(MapValueForKey(gTextures, "plasmavial"))
		plasmaammo.amount = 50
		plasmaammo.weapon = gWeaponParusu
		plasmaammo.obituary = "GOT SOME VIALS OF PLASMA"
		AddEntity(plasmaammo)
	End Method
	Method Update2()
		If pawn = True Then 
			spriteAngle = 0
			baseTextureName = "PAWN"
		EndIf
		If state <> 2 And reactionTimer > 25 Then
			angle = ATan2(target.z-z, target.x-x)
			If angle >= 360.0 Then angle :- 360.0
			If angle < 0.0 Then angle :+ 360.0
			dx=0.0
			dz=0.0
			If firing = False Then
				stateTimer :- 1
				If stateTimer <= 0 Then
					stateTimer = 50
					firing = True
					animation = fireAnimation
					animationRepeat = True
					anim = 0
					frame = 0
				EndIf
			Else
				stateTimer :- 1
				If stateTimer <= 0 Then
					stateTimer = 75
					firing = False
					animation = seeAnimation
					animationRepeat = False
					anim = 0 
					frame = 1
				EndIf
				fireTimer :+ 1
				If fireTimer > fireSpeed Then
					fireTimer = 0
					fireball = New TFireball
					fireball.x = x
					fireball.y = 0.25
					fireball.z = z
					fireball.direction = angle
					fireball.speed = 0.65
					fireball.damage = 10
					fireball.texture.name = "PLASMABALL_"
					AddEntity(fireball)
					Play3DSound("fireball", x, z, 40.0)
				EndIf
			EndIf
		Else
			reactionTimer :+ 1
		EndIf
	End Method
	Method Render()
		If pawn = True And active = False Then texture = TTexture(MapValueForKey(gTextures, "PAWN0_0"))
		If texture <> Null Then glBindTexture GL_TEXTURE_2D, texture.id
		glMatrixMode(GL_MODELVIEW)
		glPushMatrix()
		glTranslatef x, y, z
		glRotatef yaw, 0, 1, 0
		
		If flipSprite = False Then
			glCallList(gDisplayLists[1])
		Else
			glCallList(gDisplayLists[5])
		EndIf
		
		glLoadIdentity()
		glMatrixMode(GL_MODELVIEW)
		glPopMatrix()
		glBindTexture GL_TEXTURE_2D, Null
	End Method
EndType

Type TCacoWraith Extends TMonster
	Field firing:Byte = False
	Field fireAnimation:Byte[]
	Field timer:Int = 0
	Field cacoball:TCacoBall
	Field m:TMonster = Null
	Method New()
		class = "monster"
		radius = 0.5
		baseTextureName = "CACOWRAITH"
		health = 85
		If gGameDifficulty = DIFFICULTY_EASIEST Then
			health = 65
		EndIf
		maxHealth = health
		animation = New Byte[4]
		animation[0] = 0
		animation[1] = 1
		animation[2] = 2
		animation[3] = 3
		walkAnimation = animation
		deathAnimation = New Byte[5]
		deathAnimation[0] = 7
		deathAnimation[1] = 8
		deathAnimation[2] = 9
		deathAnimation[3] = 10
		deathAnimation[4] = 11
		fireAnimation = New Byte[2]
		fireAnimation[0] = 4
		fireAnimation[1] = 5
		hurtFrame = 6
		speed = 0.08
		seeSound = TSound(MapValueForKey(gSounds, "cacowraith0"))
		painSound = TSound(MapValueForKey(gSounds, "cacowraith1"))
		deathSound = TSound(MapValueForKey(gSounds, "cacowraith2"))
		frame = 0
		anim = 0
		texture = TTexture(MapValueForKey(gTextures, "CACOWRAITH0_0"))
		contactDamage = 2
		target = gPlayer
		cwiw = False
		canConveyor = False
	End Method
	Method Activate()
		If foundWall = False Then
			active = True
			MonsterPlaySound(seeSound)
			angle = ATan2(target.z-z, target.x-x)
			If angle >= 360.0 Then angle :- 360.0
			If angle < 0.0 Then angle :+ 360.0
			gPlayer.hearRange = 30.0
		EndIf
	End Method
	Method Update2()
		If state <> 2 And state <> 1 Then
			If Not firing Then
				A_WraithChase()
				For e = EachIn gEntities
					If e.class = "corpse" And Distance(e.x, e.z, x, z)-4.0-e.radius <= 0.0 And e <> Self Then
						m = TMonster(e)
						If m.state = 2 Then m.Revive()
					EndIf
				Next
				animation = walkAnimation
				animSpeed = 10
				timer :+ 1
				If timer > 50 Then
					firing = True
					timer = 0
					animation = fireAnimation
					animSpeed = 5
					frame = 0
					anim = 0
					animationRepeat = False
				EndIf
			Else
				dx = 0.0
				dz = 0.0
				angle = ATan2(target.z-z, target.x-x)
				timer :+ 1
				If timer = 10 Then
					cacoball = New TCacoBall
					cacoball.x=x
					cacoball.y=y
					cacoball.z=z
					cacoball.direction = angle
					cacoball.owner = Self
					AddEntity(cacoball)
				EndIf
				If timer > 50 Then
					firing = False
					timer = 0
					animation = walkAnimation
					animSpeed = 10
					frame = 0
					anim = 0
					animationRepeat = True
				EndIf
			EndIf
		EndIf
	End Method
EndType

Type TPrisrak Extends TMonster
	Field fireAnimation:Byte[]
	Field teleportAnimation:Byte[]
	Field timer:Int = 0
	Field timer2:Int = 0
	Field fireball:TFireball
	Method New()
		class = "monster"
		radius = 0.5
		baseTextureName = "PRISRAK"
		health = 60
		If gGameDifficulty = DIFFICULTY_EASIEST Then
			health = 40
		EndIf
		maxHealth = health
		animation = New Byte[4]
		animation[0] = 0
		animation[1] = 1
		animation[2] = 2
		animation[3] = 3
		walkAnimation = animation
		deathAnimation = New Byte[4]
		deathAnimation[0] = 11
		deathAnimation[1] = 12
		deathAnimation[2] = 13
		deathAnimation[3] = 14
		fireAnimation = New Byte[2]
		fireAnimation[0] = 4
		fireAnimation[1] = 5
		teleportAnimation = New Byte[7]
		teleportAnimation[0] = 6
		teleportAnimation[1] = 7
		teleportAnimation[2] = 8
		teleportAnimation[3] = 9
		teleportAnimation[4] = 8
		teleportAnimation[5] = 7
		teleportAnimation[6] = 6
		hurtFrame = 10
		speed = 0.08
		seeSound = TSound(MapValueForKey(gSounds, "prisrak0"))
		painSound = TSound(MapValueForKey(gSounds, "prisrak1"))
		deathSound = TSound(MapValueForKey(gSounds, "prisrak2"))
		frame = 0
		painChance = 5
		anim = 0
		texture = TTexture(MapValueForKey(gTextures, "PRISRAK0_0"))
		contactDamage = 2
		target = gPlayer
		cwiw = False
		canConveyor = False
	End Method
	Method Activate()
		If foundWall = False Then
			active = True
			MonsterPlaySound(seeSound)
			angle = ATan2(target.z-z, target.x-x)
			If angle >= 360.0 Then angle :- 360.0
			If angle < 0.0 Then angle :+ 360.0
			gPlayer.hearRange = 30.0
		EndIf
	End Method
	Method Update2()
		If state <> 2 And state <> 1 Then
			If timer < 100 Then
				A_WraithChase()
				timer :+ 1
			ElseIf timer = 100 Then
				timer :+ 1
				animation = fireAnimation
				dx = 0.0
				dz = 0.0
				angle = ATan2(target.z-z, target.x-x)
				If angle >= 360.0 Then angle :- 360.0
				If angle < 0.0 Then angle :+ 360.0
				frame = 0
				anim = 0
				animationRepeat = False
				animSpeed = 3
			ElseIf timer < 142 And timer > 100
				angle = ATan2(target.z-z, target.x-x)
				If angle >= 360.0 Then angle :- 360.0
				If angle < 0.0 Then angle :+ 360.0
				timer :+ 1
				timer2 :+ 1
				If timer2 > 10 Then
					timer2 = 0
					fireball = New TFireball
					fireball.x = x
					fireball.y = 0.0
					fireball.z = z
					fireball.direction = angle + (Rand(-1, 1) * 10)
					fireball.speed = 0.33
					fireball.damage = 15
					fireball.texture.name = "FIREBALL_"
					AddEntity(fireball)
					Play3DSound("fireball", x, z, 40.0)
				EndIf
				If timer >= 141 Then
					timer = 142
					timer2 = 0
					animation = teleportAnimation
					anim = 0
					frame = 0
					animationRepeat = False
					animSpeed = 5
				EndIf
			EndIf
			If timer >= 142 Then
				timer :+ 1
				For Local e:TEntity = EachIn gEntities
					If e <> Self And e.class = "secretwall" Then
						If x+radius>=e.x-1.0 And x-radius<=e.x+1.0 And z+radius>=e.z-1.0 And z-radius<=e.z+1.0 Then
							timer = 162
						EndIf
					EndIf
				Next
				dx = 0.0
				dz = 0.0
				If timer = 162 Then
					dx = GetCos(angle+90.0)*4.0
					dz = GetSin(angle+90.0)*4.0
					If CheckWall(x+dx, z+dz, radius, Self, False, True) Then
						dx = GetCos(angle-90.0)*4.0
						dz = GetSin(angle-90.0)*4.0
						If CheckWall(x+dx, z+dz, radius, Self, False, True) Then
							dx = 0.0
							dz = 0.0
						EndIf
					EndIf
					If Distance(gPlayer.x, gPlayer.z, x+dx, z+dz)-radius-gPlayer.radius <= 0 Then
						dx = 0.0
						dz = 0.0
					EndIf
				EndIf
				If timer >= 178 Then
					timer = 0
					animation = walkAnimation
					animationRepeat = True
					anim=0
					frame =0
					timer2 = 0
					animSpeed = 10
				EndIf
			EndIf
		EndIf
	End Method
EndType

Type TMissile Extends TProjectile
	Field spriteAngle:Byte = 0
	Method New()
		class = "projectile"
		friendly = False
		radius = 0.25
		texture = New TTexture
		texture.name = "missile0_"
		speed = 0.3
		damage = 20
		maxFrames = 2
	End Method
	Method Update()
		spriteAngle = GetSpriteAngle(x, z, direction)
		flipSprite = False
		If spriteAngle = 3 Then 
			spriteAngle = 1
			flipSprite = True
		EndIf
		texture.name = "missile"+spriteAngle+"_"
		A_Projectile()
	End Method
	Method OnDestroy()
		Explode(x, y, z, 2.5, damage, direction, owner)
	End Method
End Type

Type TBanshee Extends TMonster
	Field missile:TMissile
	Field firing:Byte = False
	Field timer:Int = 35
	Field fireAnimation:Byte[]
	Method New()
		class = "monster"
		radius = 0.9
		baseTextureName = "BANSHEE"
		health = 140
		If gGameDifficulty = DIFFICULTY_EASIEST Then
			health = 110
		EndIf
		maxHealth = health
		animation = New Byte[4]
		animation[0] = 0
		animation[1] = 1
		animation[2] = 0
		animation[3] = 2
		walkAnimation = animation
		deathAnimation = New Byte[4]
		deathAnimation[0] = 6
		deathAnimation[1] = 7
		deathAnimation[2] = 8
		deathAnimation[3] = 9
		hurtFrame = 5
		fireAnimation = New Byte[2]
		fireAnimation[0] = 3
		fireAnimation[1] = 4
		speed = 0.06
		frame = 0
		anim = 0
		seeSound = TSound(MapValueForKey(gSounds, "banshee0"))
		painSound = TSound(MapValueForKey(gSounds, "banshee1"))
		deathSound = TSound(MapValueForKey(gSounds, "banshee2"))
		texture = TTexture(MapValueForKey(gTextures, "BANSHEE0_0"))
		contactDamage = 1
		target = gPlayer
		painChance = 10
	End Method
	Method Activate()
		If foundWall = False Then
			active = True
			angle = ATan2(target.z-z, target.x-x)
			If angle >= 360.0 Then angle :- 360.0
			If angle < 0.0 Then angle :+ 360.0
			gPlayer.hearRange = 30.0
			MonsterPlaySound(seeSound)
		EndIf
	End Method
	Method Update2()
		If state = 0 Then
			If Not firing Then
				A_WraithChase()
				timer :+ 1
				If timer > 80 Then
					timer = 0
					anim = 0
					frame = 0
					animation = fireAnimation
					animationRepeat = True
					animSpeed = 10
					firing = True
					dx = 0.0
					dz = 0.0
				EndIf
			Else
				angle = ATan2((gPlayer.z+gPlayer.dz)-z, (gPlayer.x+gPlayer.dx)-x)
				timer :+ 1
				If timer = 20 Or timer = 40 Or timer = 59 Then
					missile= New TMissile
					missile.x = x + GetCos(angle) * 0.1 + GetCos(angle + 90.0) * 0.5
					missile.z = z + GetSin(angle) * 0.1 + GetSin(angle + 90.0) * 0.5
					missile.y = y - 0.125
					missile.direction = angle
					missile.friendly = False
					missile.owner = Self
					AddEntity(missile)
					missile = New TMissile
					missile.x = x + GetCos(angle) * 0.1 + GetCos(angle - 90.0) * 0.5
					missile.z = z + GetSin(angle) * 0.1 + GetSin(angle - 90.0) * 0.5
					missile.y = y - 0.125
					missile.direction = angle
					missile.friendly = False
					missile.owner = Self
					AddEntity(missile)
					Play3DSound("doublegrenadelaunch", x, z)
				EndIf
				If timer > 60 Then
					timer = 0
					anim = 0
					frame = 0
					animation = walkAnimation
					animationRepeat = True
					animSpeed = 10
					firing = False
				EndIf
			EndIf
		EndIf
	End Method
End Type

Type TMutant Extends TMonster
	Field fireAnimation:Byte[]
	Field fireAnimation2:Byte[]
	Field firing:Byte = False
	Field fireball:TFireball
	Field timer:Int = 60
	Method New()
		class = "monster"
		radius = 0.5
		baseTextureName = "MUTANT"
		health = 90
		If gGameDifficulty = DIFFICULTY_EASIEST Then
			health = 70
		EndIf
		maxHealth = health

		animation = New Byte[4]
		animation[0] = 0
		animation[1] = 1
		animation[2] = 0
		animation[3] = 2
		walkAnimation = animation
		deathAnimation = New Byte[3]
		deathAnimation[0] = 8
		deathAnimation[1] = 9
		deathAnimation[2] = 10
		hurtFrame = 7
		speed = 0.09
		frame = 0
		anim = 0
		texture = TTexture(MapValueForKey(gTextures, "MUTANT0_0"))
		contactDamage = 1
		target = gPlayer
		seeSound = TSound(MapValueForKey(gSounds, "mutant0"))
		painSound = TSound(MapValueForKey(gSounds, "mutant1"))
		deathSound = TSound(MapValueForKey(gSounds, "mutant2"))
		fireAnimation = New Byte[2]
		fireAnimation[0] = 3 
		fireAnimation[1] = 4
		fireAnimation2 = New Byte[2]
		fireAnimation2[0] = 5
		fireAnimation2[1] = 6
		animSpeed = 5
		animationRepeat = True
	End Method
	Method Update2()
		If state = 0 Then
			If firing = False Then 
				A_WraithChase()
				timer :+ 1
				If timer > 100 Then
					timer = 0
					firing = True
					animation = fireAnimation
					animSpeed = 10
					anim = 0
					frame = 0
					dx = 0.0
					dz = 0.0
					spriteAngle = 0
					angle = ATan2(gPlayer.z - z, gPlayer.x-x)
				EndIf
			Else
				angle = ATan2(gPlayer.z - z, gPlayer.x-x)
				spriteAngle = 0
				timer :+ 1
				
				If timer = 21 Then
					animation = fireAnimation2
					anim = 0
					frame = 0
					animSpeed = 5
				EndIf
				
				If timer >= 20 And timer Mod 10 = 0 Then
					For Local i:Int = 0 To 2
						fireball = New TFireball
						fireball.x = x
						fireball.z = z
						fireball.direction = angle - 9.0 + (9.0 * i)
						fireball.speed = 0.425
						fireball.damage = 15
						AddEntity(fireball)
						Play3DSound("fireball", x, z, 40.0)
					Next
				EndIf
				
				If timer > 70 Then
					timer = 0
					firing = False
					animation = walkAnimation
					anim = 0
					frame = 0
				EndIf
			EndIf
		EndIf
	End Method
EndType

Type TFundie Extends TMonster
	Field fireball:TFireball
	Field firing:Byte = False
	Field timer:Int = 0
	Field fireAnimation:Byte[]
	Field fireAnimation2:Byte[]
	Method New()
		class = "monster"
		radius = 0.5
		baseTextureName = "FUNDIE"
		maxHealth = 55
		health = 55
		animation = New Byte[4]
		animation[0] = 0
		animation[1] = 1
		animation[2] = 2
		animation[3] = 1
		walkAnimation = animation
		deathAnimation = New Byte[1]
		deathAnimation[0] = 6
		fireAnimation = New Byte[1]
		fireAnimation[0] = 3
		fireAnimation2 = New Byte[1]
		fireAnimation2[0] = 4
		hurtFrame = 5
		makeNoise = False
		speed = 0.08
		frame = 0
		anim = 0
		texture = TTexture(MapValueForKey(gTextures, "FUNDIE0_0"))
		contactDamage = 1
		target = gPlayer
	End Method
	Method Activate()
		If foundWall = False Then
			active = True
			angle = ATan2(target.z-z, target.x-x)
			If angle >= 360.0 Then angle :- 360.0
			If angle < 0.0 Then angle :+ 360.0
			gPlayer.hearRange = 30.0
		EndIf
	End Method
	Method OnDie()
		Explode(x, y, z, 1.0, 10)
	EndMethod
	Method Update2()
		If state <> 2 And state <> 1 Then
			If Not firing Then
				timer :+ 1
				A_WraithChase()
				If timer > 50 Then 
					firing = True
					timer = 0
					dx = 0.0
					dz = 0.0
					angle = ATan2(gPlayer.z-z, gPlayer.x-x)
					fireball = New TFireball
					fireball.x = x + GetCos(angle + 90.0) * 0.5
					fireball.z = z + GetSin(angle + 90.0) * 0.5
					fireball.y = y
					fireball.damage = 10
					fireball.speed = 0.35
					fireball.texture = New TTexture
					fireball.texture.name = "FUNDIEHAND_"
					fireball.maxFrames = 0
					fireball.direction = angle
					AddEntity(fireball)
					fireball = New TFireball
					fireball.x = x + GetCos(angle - 90.0) * 0.5
					fireball.z = z + GetSin(angle - 90.0) * 0.5
					fireball.y = y
					fireball.damage = 10
					fireball.speed = 0.35
					fireball.texture = New TTexture
					fireball.texture.name = "FUNDIEHAND_"
					fireball.maxFrames = 0
					fireball.direction = angle
					AddEntity(fireball)
					animation = fireAnimation
					anim = 0
					frame = 0
					Play3DSound("grenadelaunch", x, z)
				EndIf
			Else
				angle = ATan2(gPlayer.z-z, gPlayer.x-x)
				timer :+ 1
				If timer > 40 Then
					animation = fireAnimation2
					anim = 0
					frame = 0
				EndIf
				If timer > 50 Then
					firing = False
					timer = 0
					animation = walkAnimation
					anim = 0
					frame = 0
				EndIf
			EndIf
		EndIf
	End Method
End Type

Type TMecha Extends TMonster
	Field mutant:TMutant
	Field missile:TMissile
	Field phase:Int = 0
	Field phaseTimer:Int = 0
	Field fireTimer:Int = 0
	Field fireAnimation:Byte[]
	Field fireAnimation2:Byte[]
	Field fireAnimation3:Byte[]
	Field chargeAnimation:Byte[]
	Field fireball:TFireball
	'7 8 9 10 11
	Method New()
		class = "monster"
		radius = 1.0
		baseTextureName = "MECHA"
		health = 2250
		If gGameDifficulty = DIFFICULTY_EASY Then
			health = 2100
		EndIf
		If gGameDifficulty = DIFFICULTY_EASIEST Then
			health = 1900
		EndIf
		maxHealth = health
		animation = New Byte[4]
		animation[0] = 0
		animation[1] = 1
		animation[2] = 2
		animation[3] = 3
		walkAnimation = animation
		deathAnimation = New Byte[5]
		deathAnimation[0] = 14
		deathAnimation[1] = 15
		deathAnimation[2] = 16
		deathAnimation[3] = 17
		deathAnimation[4] = 18
		fireAnimation = New Byte[3]
		fireAnimation[0] = 4
		fireAnimation[1] = 5
		fireAnimation[2] = 6
		fireAnimation2 = New Byte[3]
		fireAnimation2[0] = 7
		fireAnimation2[1] = 8
		fireAnimation2[2] = 9
		fireAnimation3 = New Byte[2]
		fireAnimation3[0] = 10
		fireAnimation3[1] = 11
		chargeAnimation = New Byte[2]
		chargeAnimation[0] = 12
		chargeAnimation[1] = 13
		hurtFrame = 0
		speed = 0.08
		seeSound = TSound(MapValueForKey(gSounds, "providence0"))
		painSound = TSound(MapValueForKey(gSounds, "cacowraith1"))
		deathSound = TSound(MapValueForKey(gSounds, "providence1"))
		frame = 0
		painChance = -1
		anim = 0
		texture = TTexture(MapValueForKey(gTextures, "MECHA0_0"))
		contactDamage = 2
		target = gPlayer
		makeNoise = True
	End Method
	Method Activate()
		If foundWall = False And active = False Then
			active = True
			MonsterPlaySound(seeSound)
			angle = ATan2(target.z-z, target.x-x)
			If angle >= 360.0 Then angle :- 360.0
			If angle < 0.0 Then angle :+ 360.0
			gPlayer.hearRange = 30.0
			StopChannel(gMusicChannel)
			gMusicChannel = PlaySound(gBossMusic)
		EndIf
	End Method
	Method Update2()
		If state <> 2 Then
			phaseTimer :+ 1
			If phaseTimer > 100 Then
				phaseTimer = 0
				If phase = 0 Then 
					phase = Rand(1, 3)
				Else
					phase = 0
				EndIf
				frame = 0
				anim = 0
				fireTimer = 0
				Select phase
					Case 0
						animation = walkAnimation
						animationRepeat = True
						animSpeed = 10
					Case 1
						animation = fireAnimation
						animationRepeat = True
						animSpeed = 5
						dx = 0.0
						dz = 0.0
					Case 2
						animation = fireAnimation2
						animationRepeat = True
						animSpeed = 5
						dx = 0.0
						dz = 0.0
					Case 3
						animation = fireAnimation2
						animationRepeat = True
						animSpeed = 5
						dx = 0.0
						dz = 0.0
				EndSelect
			EndIf
		EndIf
		If state = 0 Then
			If phase = 0 Then
				A_WraithChase()
			ElseIf phase = 1
				spriteAngle = 0
				angle = ATan2(gPlayer.z-z, gPlayer.x-x)
				fireTimer :+ 1
				If fireTimer = 15 Then
					fireTimer = 0
					missile= New TMissile
					missile.x = x + GetCos(angle) * 0.1 + GetCos(angle + 90.0) * 0.5
					missile.z = z + GetSin(angle) * 0.1 + GetSin(angle + 90.0) * 0.5
					missile.y = y - 0.125
					missile.direction = angle + Rnd(-20.0, 20.0)
					missile.friendly = False
					missile.speed = 0.35
					missile.owner = Self
					AddEntity(missile)
					missile = New TMissile
					missile.x = x + GetCos(angle) * 0.1 + GetCos(angle - 90.0) * 0.5
					missile.z = z + GetSin(angle) * 0.1 + GetSin(angle - 90.0) * 0.5
					missile.y = y - 0.125
					missile.direction = angle + Rnd(-20.0, 20.0)
					missile.friendly = False
					missile.owner = Self
					missile.speed = 0.35
					AddEntity(missile)
					Play3DSound("doublegrenadelaunch", x, z)
				EndIf
			ElseIf phase = 2 Then
				spriteAngle = 0
				angle = ATan2(gPlayer.z-z, gPlayer.x-x)
				fireTimer :+ 1
				If fireTimer = 15 Then
					animation = fireAnimation3
					frame = 0
					anim = 0
					animationRepeat = True
					animSpeed = 5
				EndIf
				If fireTimer = 25 Then
					fireTimer = 16
					For Local i:Int = 0 To 7
						fireball = New TFireball
						fireball.x = x
						fireball.z = z
						fireball.direction = 45.0 * i
						fireball.speed = 0.425
						fireball.damage = 15
						AddEntity(fireball)
						Play3DSound("fireball", x, z, 40.0)
					Next
				EndIf
			ElseIf phase = 3
				spriteAngle = 0
				angle = ATan2(gPlayer.z-z, gPlayer.x-x)
				fireTimer :+ 1
				If fireTimer = 15 Then
					animation = chargeAnimation
					frame = 0
					anim = 0
					animationRepeat = True
					animSpeed = 5
					dx = Cos(angle) * 0.5
					dz = Sin(angle) * 0.5
					Play3DSound("boost", x, z)
				EndIf
				If CheckWall(x+dx, z+dz, radius, Self) Then
					dx = 0.0
					dz = 0.0
					fireTimer = 0
					animation = fireAnimation3
					frame = 0
					anim = 0
					animationRepeat = True
					animSpeed = 5				
				EndIf
				If Distance(gPlayer.x, gPlayer.z, x, z)-radius-gPlayer.radius <= 0.0 And fireTimer > 15 Then
					knockback = 0.5
					knockbackDirection = angle + 180.0
					gPlayer.health :- 20
				EndIf
			EndIf
		EndIf
	End Method
	Method OnDie()
		ActivateAllWithId(10)
		mutant = New TMutant
		mutant.x = x
		mutant.z = z
		mutant.angle = ATan2(gPlayer.z-z, gPlayer.x-x)
		AddEntity(mutant)
		Play3DSound("explosion", x, z)
	End Method
EndType

Type TGod Extends TMonster
	Field fireAnimation:Byte[]
	Field fireTimer:Int = 0
	Field firing:Int = False
	Field fireball:TFireball
	Field velocity:Float = 0.0
	Field jumpAnimation:Byte[]
	Field f:Int = False
	Field timer:Float = 0.0
	Method New()
		class = "monster"
		radius = 0.5
		baseTextureName = "GOD"
		maxHealth = 10
		health = 10
		speed = 0.1
		painChance = -1
		frame = 0
		anim = 0
		seeSound = TSound(MapValueForKey(gSounds, "cacowraith0"))
		painSound = TSound(MapValueForKey(gSounds, "cacowraith1"))
		deathSound = TSound(MapValueForKey(gSounds, "god1"))
		makeNoise = True
		target = gPlayer
		animation = New Byte[4]
		animation[0] = 0
		animation[1] = 1
		animation[2] = 0
		animation[3] = 2
		walkAnimation = animation
		deathAnimation = New Byte[5]
		deathAnimation[0] = 4
		deathAnimation[1] = 5
		deathAnimation[2] = 6
		deathAnimation[3] = 7
		deathAnimation[4] = 8
		fireAnimation = New Byte[2]
		fireAnimation[0] = 9
		fireAnimation[1] = 3
		jumpAnimation = New Byte[1]
		jumpAnimation[0] = 1
		cwiw = False
		y = -16.0
	End Method
	Method Activate()
		active = True
		angle = ATan2(target.z-z, target.x-x)
		If angle >= 360.0 Then angle :- 360.0
		If angle < 0.0 Then angle :+ 360.0
		gPlayer.hearRange = 30.0
		state = 3
		timer = -1
		anim = 0
		frame = 0
		velocity = 2.0
		animation = jumpAnimation
		spriteAngle = 0
	End Method
	Method Update2()
		spriteAngle = 0
		If state = 0 Then
			If Not firing Then
				A_WraithChase()
				fireTimer :+ 1
				If fireTimer > 100 Then
					fireTimer = 0
					firing = True
					animation = fireAnimation
					fireball = New TFireball
					fireball.x = x
					fireball.z = z
					fireball.damage = 1
					fireball.direction = ATan2(gPlayer.z-z, gPlayer.x-x)
					fireball.friendly = False
					fireball.owner = Self
					AddEntity(fireball)
					animationRepeat = False
					animSpeed = 3
					anim = 0
					frame = 0
					Play3DSound("fireball", x, z)
				EndIf
			Else
				dx = 0.0
				dz = 0.0
				angle = ATan2(gPlayer.z-z, gPlayer.x-x)
				fireTimer :+ 1
				If fireTimer > 50 Then
					fireTimer = 0
					firing = False
					animation = walkAnimation
					frame = 0
					anim = 0
					
					animationRepeat = True
					animSpeed = 10
				EndIf
			EndIf
		ElseIf state = 3 Then
			spriteAngle = 0
			
			angle = ATan2(gPlayer.z-z, gPlayer.x-x)
			y :+ velocity
			velocity :- 0.05
			
			If y > 0.0 And velocity > 0.0 Then f = True
			
			If y <= 0.0 And f = True Then
				y = 0.0
				velocity = 0.0
				state = 0
				animation = walkAnimation
				anim = 0
				frame = 0
			EndIf
			x:+GetCos(angle) * speed * 2.0
			z:+GetSin(angle) * speed * 2.0
			texture = TTexture(MapValueForKey(gTextures, "GOD0_1"))
		EndIf
	End Method
	Method OnDie()
		ActivateAllWithId(10)
	End Method
End Type

Type TDemon Extends TMonster
	Field startAnimation:Byte[]
	Field introAnimation:Byte[]
	Field idleAnimation:Byte[]
	Field fireAnimation:Byte[]
	Field hurtAnimation:Byte[]
	Field timer:Int = -1
	Field explosionTimer:Int = 0
	Field phase:Int = 0
	Field phaseTimer = 0
	Field fireball:TFireball
	Field nospam:Byte = False
	Method New()
		class = "monster"
		radius = 4.0
		baseTextureName = "DEMON"
		health = 3500
		If gGameDifficulty = DIFFICULTY_EASY Then
			health = 3100
		EndIf
		If gGameDifficulty = DIFFICULTY_EASIEST Then
			health = 2900
		EndIf
		maxHealth = health
		speed = 0.0
		painChance = -1
		frame = 0
		anim = 0
		seeSound = TSound(MapValueForKey(gSounds, "demon0"))
		painSound = TSound(MapValueForKey(gSounds, "cacowraith1"))
		deathSound = TSound(MapValueForKey(gSounds, "demon1"))
		texture = TTexture(MapValueForKey(gTextures, "DEMON0_0"))
		startAnimation = New Byte[1]
		startAnimation[0] = 0
		introAnimation = New Byte[7]
		introAnimation[0] = 0
		introAnimation[1] = 1
		introAnimation[2] = 2
		introAnimation[3] = 3
		introAnimation[4] = 4
		introAnimation[5] = 5
		introAnimation[6] = 6
		idleAnimation = New Byte[1]
		idleAnimation[0] = 6
		fireAnimation = New Byte[3]
		fireAnimation[0] = 7
		fireAnimation[1] = 8
		fireAnimation[2] = 7
		hurtAnimation = New Byte[2]
		hurtAnimation[0] = 9
		hurtAnimation[1] = 10
		animation = startAnimation
		target = gPlayer
		makeNoise = True
		y = -16.0
		bleed = False
		deathAnimation = New Byte[1]
		deathAnimation[0] = 0
	End Method
	Method Activate()
		If foundWall = False And active = False Then
			active = True
			angle = ATan2(target.z-z, target.x-x)
			If angle >= 360.0 Then angle :- 360.0
			If angle < 0.0 Then angle :+ 360.0
			gPlayer.hearRange = 30.0
			state = 3
			timer = -1
			anim = 0
			frame = 0
		EndIf
	End Method
	Method Update2()
		spriteAngle = 0
		dx = 0.0
		dz = 0.0
		knockback = 0.0
		If state = 3 Then
			health = maxHealth
			gCameraShakeTimer = 1
			y :+ 0.05
			If y >= 4.0 Then 
				y = 4.0
				If timer = -1 Then
					animation = introAnimation
					animationRepeat = False
					animSpeed = 5
					anim = 0
					frame = 0
					timer = 0
					MonsterPlaySound(seeSound)
				EndIf
			EndIf
			If timer > -1 Then
				timer :+ 1
				If timer > 160 Then
					timer = 0
					animation = idleAnimation
					animationRepeat = True
					animSpeed = 5
					anim = 0
					frame = 0
					state = 0
					StopChannel(gMusicChannel)
					gMusicChannel = PlaySound(gFinalBossMusic)
				EndIf
			EndIf
			
			anim :+ 1
			If anim > animSpeed Then
				anim = 0
				frame :+ 1
				If frame > Len(animation) - 1 
					If animationRepeat = True Then 
						frame = 0
					Else
						frame = Len(animation) - 1
					EndIf
				EndIf
			EndIf
			texture = TTexture(MapValueForKey(gTextures, baseTextureName + spriteAngle + "_" + animation[frame]))
		EndIf
		If state = 0 Then
			
			phaseTimer :+ 1
			If phaseTimer > 100 Then
				phaseTimer = 0
				phase = Rand(1, 2)
				If Rand(0, 5) = 4 Then 
					If nospam = False Then
						phase = 0
						nospam = True
					EndIf
				Else
					nospam = False
				EndIf
				frame = 0
				anim = 0
				Select phase
					Case 0
						animation = idleAnimation
						animationRepeat = False
						animSpeed = 5
						'x 640
						'ys 208 368
						For i = 0 To 1
							Local fireWraith:TFireWraith = New TFireWraith
							fireWraith.x = 80.5
							fireWraith.z = 26.5
							fireWraith.angle = ATan2(gPlayer.z-fireWraith.z, gPlayer.x-fireWraith.x)
							
							If i = 1 Then fireWraith.z = 46.5
							ListAddLast gMonsters, fireWraith
							gTotalMonsters :+ 1
							For Local m:TMonster = EachIn gMonsters
								If m.class = "monster" And Distance(m.x, m.z, fireWraith.x, fireWraith.z)-fireWraith.radius*2 < 0.0 And m <> fireWraith Then
									ListRemove gMonsters, fireWraith
									gTotalMonsters :- 1
								EndIf 
							Next
						Next
					Case 1
						animation = fireAnimation
						animationRepeat = True
						animSpeed = 5
						timer = 0
					Case 2
						animation = fireAnimation
						animationRepeat = True
						animSpeed = 5
						timer = 0
						angle = 90
				End Select
			EndIf
			If phase = 1 Then
				angle = ATan2(gPlayer.z-z, gPlayer.x-x)
				timer :+ 1
				If timer > 50 Then
					For i = 0 To 4
						fireball = New TFireball
						fireball.x = x
						fireball.z = z
						fireball.friendly = False
						fireball.damage = 10
						fireball.owner = Self
						fireball.direction = (angle-25 + Rand(-25, 25)) + (i*10)
						AddEntity(fireball)
						fireball.speed = 0.43
					Next
					timer = 0
					Play3DSound("fireball", x, z, 40.0)
				EndIf
			ElseIf phase = 2 Then
				angle :+ 2.0
				timer :+ 1
				If timer > 5 Then
					fireball = New TFireball
					fireball.x = x
					fireball.z = z
					fireball.damage = 10
					fireball.owner = Self
					fireball.direction = angle
					fireball.texture.name = "PLASMABALL_"
					Play3DSound("parusu", x, z, 40.0)
					AddEntity(fireball)
					timer = 0
				EndIf
			EndIf
		EndIf
		If state = 4 Then
			y :- 0.05
			explosionTimer :+ 1
			If explosionTimer > 10 Then
				explosionTimer = 0
				Explode(x+Rnd(-4, 4), y, z+Rnd(-4, 4), 0.0, 0.0, 0.0)
			EndIf
			anim :+ 1
			If anim > animSpeed Then
				anim = 0
				frame :+ 1
				If frame > Len(animation) - 1 
					If animationRepeat = True Then 
						frame = 0
					Else
						frame = Len(animation) - 1
					EndIf
				EndIf
			EndIf
			
			texture = TTexture(MapValueForKey(gTextures, baseTextureName + "0_" + animation[frame]))
			If y < -16.0 Then
				ListRemove gMonsters, Self
				Local god:TGod = New TGod
				god.x = x
				god.z = z
				god.angle = ATan2(gPlayer.z-z, gPlayer.x-x)
				ListAddLast gMonsters, god
			EndIf
		EndIf
	End Method
	Method OnDie()
		state = 4
		anim = 0
		frame = 0
		animSpeed = 3
		animationRepeat = True
		animation = hurtAnimation
		For Local m:TMonster = EachIn gMonsters
			If m.class = "monster" Then
				m.health = 0
			EndIf
		Next
		StopChannel(gMusicChannel)
		SetChannelVolume(PlaySound(deathSound), sSoundVolume * 2.0)
	End Method
	Method Render()
		If texture <> Null Then glBindTexture GL_TEXTURE_2D, texture.id
		glMatrixMode(GL_MODELVIEW)
		glPushMatrix()
		glTranslatef x, y, z
		glRotatef yaw, 0, 1, 0
		glScalef 8.0, 8.0, 8.0
		
		glCallList(gDisplayLists[1])
		
		glLoadIdentity()
		glMatrixMode(GL_MODELVIEW)
		glPopMatrix()
		glBindTexture GL_TEXTURE_2D, Null
	End Method
EndType

Type TProvidence Extends TMonster
	Field deathAnimation2:Byte[]
	Field missileAnimation:Byte[]
	Field plasmaAnimation:Byte[]
	Field explosionTimer:Int = 0
	Field deathTimer:Int = 0
	Field phase:Int = 0
	Field phaseTimer:Int = 0
	Field fireTimer:Int = 0
	Field grenadeRate:Int = 20
	
	Field grenade:TGrenade
	Field plasma:THomingBall
	Field prop:TProp
	Method New()
		class = "monster"
		radius = 1.5
		baseTextureName = "PROVIDENCE"
		health = 2000
		If gGameDifficulty = DIFFICULTY_EASY Then
			health = 1650
			grenadeRate = 30
		EndIf
		If gGameDifficulty = DIFFICULTY_EASIEST Then
			health = 1350
			grenadeRate = 40
		EndIf
		maxHealth = health
		animation = New Byte[2]
		animation[0] = 0
		animation[1] = 1
		walkAnimation = animation
		deathAnimation2 = New Byte[2]
		deathAnimation2[0] = 6
		deathAnimation2[1] = 7
		deathAnimation = New Byte[5]
		deathAnimation[0] = 8
		deathAnimation[1] = 9
		deathAnimation[2] = 10
		deathAnimation[3] = 11
		deathAnimation[4] = 12
		missileAnimation = New Byte[3]
		missileAnimation[0] = 2
		missileAnimation[1] = 3
		missileAnimation[2] = 4
		plasmaAnimation = New Byte[2]
		plasmaAnimation[0] = 5
		plasmaAnimation[1] = 0
		hurtFrame = 0
		speed = 0.08
		seeSound = TSound(MapValueForKey(gSounds, "providence0"))
		painSound = TSound(MapValueForKey(gSounds, "cacowraith1"))
		deathSound = TSound(MapValueForKey(gSounds, "providence1"))
		frame = 0
		painChance = -1
		anim = 0
		texture = TTexture(MapValueForKey(gTextures, "PROVIDENCE0_0"))
		contactDamage = 2
		target = gPlayer
		makeNoise = True
	End Method
	Method Activate()
		If foundWall = False And active = False Then
			active = True
			MonsterPlaySound(seeSound)
			angle = ATan2(target.z-z, target.x-x)
			If angle >= 360.0 Then angle :- 360.0
			If angle < 0.0 Then angle :+ 360.0
			gPlayer.hearRange = 30.0
			StopChannel(gMusicChannel)
			gMusicChannel = PlaySound(gBossMusic)
		EndIf
	End Method
	Method Update2()
		'If health <= 0 Then health = 1
		If state = 0
			If phase = 0 Then
				A_WraithChase()
			ElseIf phase = 1 Then
				angle = ATan2(gPlayer.z-z, gPlayer.x-x)
				fireTimer :+ 1
				If fireTimer > grenadeRate Then
					fireTimer = 0
					grenade = New TGrenade
					grenade.x = x
					grenade.y = y - 0.25
					grenade.z = z
					grenade.velocity = 0.1
					grenade.speed = 1.5
					grenade.direction = angle + Rnd(-45.0, 45.0)
					grenade.friendly = False
					grenade.damage = 10
					AddEntity(grenade)
					Play3DSound("grenadelaunch", x, z)
				EndIf
			ElseIf phase = 2 Then
				angle = ATan2((gPlayer.z + gPlayer.dz)-z, (gPlayer.x + gPlayer.dx)-x)
				fireTimer :+ 1
				If fireTimer > 10 Then
					fireTimer = 0
					plasma = New THomingBall
					plasma.x = x + GetCos(angle) * 0.1 + GetCos(angle + 90.0) * 0.5
					plasma.y = y + 0.5
					plasma.z = z + GetSin(angle) * 0.1 + GetSin(angle + 90.0) * 0.5
					plasma.direction = angle
					AddEntity(plasma)
					plasma = New THomingBall
					plasma.x = x + GetCos(angle) * 0.1 + GetCos(angle - 90.0) * 0.5
					plasma.y = y + 0.5
					plasma.z = z + GetSin(angle) * 0.1 + GetSin(angle - 90.0) * 0.5
					plasma.direction = angle
					AddEntity(plasma)
					Play3dSound("fireball", x, z)
				EndIf
			EndIf
			phaseTimer :+ 1
			If phaseTimer > 100 Then
				phaseTimer = 0
				If phase = 0 Then 
					phase = Rand(1, 2)
				Else
					phase = 0
				EndIf
				frame = 0
				anim = 0
				fireTimer = 0
				Select phase
					Case 0
						animation = walkAnimation
						animationRepeat = True
						animSpeed = 10
					Case 1
						animation = missileAnimation
						animationRepeat = True
						animSpeed = 5
						dx = 0.0
						dz = 0.0
					Case 2
						animation = plasmaAnimation
						animationRepeat = True
						animSpeed = 5
						dx = 0.0
						dz = 0.0
				EndSelect
			EndIf
		ElseIf state = 3 Then
			
			explosionTimer :+ 1
			If explosionTimer > 20 Then
				explosionTimer = 0
				Explode(x+Rnd(-1, 1), y, z+Rnd(-1, 1), 0.0, 0.0, 0.0)
			EndIf
			anim :+ 1
			If anim > animSpeed Then
				anim = 0
				frame :+ 1
				If frame > Len(animation) - 1 
					If animationRepeat = True Then 
						frame = 0
					Else
						frame = Len(animation) - 1
					EndIf
				EndIf
			EndIf
			
			texture = TTexture(MapValueForKey(gTextures, baseTextureName + "0_" + animation[frame]))
			''''''''''''''''
			deathTimer :+ 1
			If deathTimer = 150 Then
				class = "corpse"
				dx = 0.0
				dz = 0.0
				animation = deathAnimation
				animationRepeat = True
				animSpeed = 10
				frame = 0
				anim = 0
				state = 2
				If makeNoise = True Then MonsterPlaySound(deathSound)
			EndIf
			'''''''''''''''
			
		EndIf
	End Method
	Method OnDie()
		
		state = 3
		animation = deathAnimation2
		animationRepeat = True
		animSpeed = 3
		ActivateAllWithId(10)
		
	End Method
	Method Render()
		If texture <> Null Then glBindTexture GL_TEXTURE_2D, texture.id
		glMatrixMode(GL_MODELVIEW)
		glPushMatrix()
		glTranslatef x, y + 1.0, z
		glRotatef yaw, 0, 1, 0
		glScalef 2.0, 2.0, 2.0
		
		If flipSprite = False Then
			glCallList(gDisplayLists[1])
		Else
			glCallList(gDisplayLists[5])
		EndIf
		
		glLoadIdentity()
		glMatrixMode(GL_MODELVIEW)
		glPopMatrix()
		glBindTexture GL_TEXTURE_2D, Null
	End Method
EndType

Type TChicken Extends TProp
	Field angle:Float = 0.0
	Field dx:Float = 0.0
	Field dz:Float = 0.0
	Field spriteAngle:Int = 0
	Field ddx:Float = 0.0
	Field ddz:Float = 0.0
	Field k:Float = 0.0
	Field a:Float = 0.0
	Field frame:Byte = 0
	Field anim:Byte = 0
	Field animation:Byte[]
	Field deathAnimation:Byte[]
	Field health:Int = 10
	Field state:Byte = 0
	
	Field e:TEntity
	Field p:TProjectile
	Method New()
		class = "chicken"
		solid = False
		radius = 0.25
		texture = TTexture(MapValueForKey(gTextures, "CHICKEN0_0"))
		animation = New Byte[4]
		animation[0] = 0
		animation[1] = 0
		animation[2] = 1
		animation[3] = 1
		deathAnimation = New Byte[4]
		deathAnimation[0] = 2
		deathAnimation[1] = 3
		deathAnimation[2] = 4
		deathAnimation[3] = 5
		angle = Rand(0, 360)
	End Method
	Method Update()
		yaw = -ATan2((gPlayer.z - gPlayer.camSin) - z, (gPlayer.x - gPlayer.camCos) - x) + 90.0
		If angle >= 360.0 Then angle :- 360.0
		If angle < 0.0 Then angle :+ 360.0
		ddx= x-gPlayer.x
		ddz= z-gPlayer.z
		k= 1.0/Sqr(ddx*ddx + ddz*ddz)
		ddx :* k
		ddz :* k
		If ddz < 0.0 Then
			a = 360.0 - ACos(ddx)
		Else
			a = ACos(ddx)
		EndIf
		a :- angle + 135.0
		If a >= 360.0 Then a :- 360.0
		If a < 0.0 Then a :+ 360.0
		spriteAngle = Abs(Floor(a)*4/360)
		flipTexture = False
		If spriteAngle = 3 Then
			spriteAngle = 1
			flipTexture = True
		EndIf
		
		If state = 0 Then
			
			For e = EachIn gEntities
				If x+radius>e.x-e.radius And x-radius<e.x+e.radius And z+radius>e.z-e.radius And z-radius<e.z+e.radius Then
					If e.class = "projectile" Then
						p = TProjectile(e)
						If TSickle(p) = Null Then RemoveEntity(e)
						health :- p.damage
						For Local i:Int = 0 To 2 
							AddParticle(x, y - 0.75, z, (RndFloat()*0.25)-0.125, 0.2, (RndFloat()*0.25)-0.125, 0.025, 20, TTexture(MapValueForKey(gTextures, "particles_1")))
						Next
						Exit
					ElseIf e.class = "monster"
						health :- TMonster(e).contactDamage
						For Local i:Int = 0 To 2 
							AddParticle(x, y - 0.75, z, (RndFloat()*0.25)-0.125, 0.2, (RndFloat()*0.25)-0.125, 0.025, 20, TTexture(MapValueForKey(gTextures, "particles_1")))
						Next
						Exit
					EndIf
				EndIf
			Next
			If CheckWall(x + GetCos(angle) * 0.1, z + GetSin(angle) * 0.1, radius, Self) Then
				angle :+ 45
			EndIf
			
			dx = GetCos(angle) * 0.1
			dz = GetSin(angle) * 0.1
			TryMove(x, z, dx, dz, radius, Self, False, False)
			x :+ dx
			z :+ dz
			
			anim :+ 1
			If anim > 5 Then
				anim = 0
				frame :+ 1
				If frame > Len(animation) - 1 
					frame = 0
				EndIf
				texture = TTexture(MapValueForKey(gTextures, "CHICKEN" + spriteAngle + "_" + animation[frame]))
				If texture = Null Then Print "CHICKEN"+spriteAngle+"_"+animation[frame]
			EndIf
			
			If health <= 0 Then
				state = 1
				animation = deathAnimation
				frame = 0
				anim = 0
				Play3DSound("chicken", x, z)
			EndIf
		Else
			anim :+ 1
			If anim > 5 Then
				anim = 0
				frame :+ 1
				If frame > 3 Then
					frame = 3
					anim = Rand(0, 8)
					Select anim
						Case 2
							Local item:TMedkit = New TMedkit
							item.x = x
							item.z = z
							item.radius = 0.5
							item.amount = 10
							item.texture = TTexture(MapValueForKey(gTextures, "stimpack"))
							AddEntity(item)
						Case 3
							Local eggammo:TAmmoPickup = New TAmmopickup
							eggammo.x = x
							eggammo.z = z
							eggammo.radius = 0.5
							eggammo.texture = TTexture(MapValueForKey(gTextures, "cartonofeggs"))
							eggammo.amount = 10
							eggammo.weapon = gWeaponChicken
							eggammo.obituary = "GOT A CARTON OF EGGS"
							AddEntity( eggammo)
						Case 5
							Local grenadeammo:TAmmoPickup = New TAmmoPickup
							grenadeammo.x = x
							grenadeammo.z = z
							grenadeammo.radius = 0.5
							grenadeammo.texture = TTexture(MapValueForKey(gTextures, "grenades"))
							grenadeammo.amount = 3
							grenadeammo.weapon = gWeaponGrenade
							grenadeammo.obituary = "GOT SOME GRENADES"
							AddEntity(grenadeammo)
						Case 7
							Local plasmaammo:TAmmoPickup = New TAmmoPickup
							plasmaammo.x = x
							plasmaammo.z = z
							plasmaammo.radius = 0.5
							plasmaammo.texture = TTexture(MapValueForKey(gTextures, "plasmavial"))
							plasmaammo.amount = 50
							plasmaammo.weapon = gWeaponParusu
							plasmaammo.obituary = "GOT SOME VIALS OF PLASMA"
							AddEntity(plasmaammo)
					EndSelect
					RemoveEntity(Self)
				EndIf
				texture = TTexture(MapValueForKey(gTextures, "CHICKEN0_" + animation[frame]))
			EndIf
		EndIf
	End Method
EndType